/**
 * (c) Atos IT Solutions and Services GmbH, 2011.
 * Licensed under EUPL 1.1
 */

package com.sag.osami.api.context;

/**
 * This interface is the central extension point of systems dealing with <em>context events</em>.
 * <p>
 * Conceptually every component which processes or generates new context events has to implement this interface and register itself in the ContextStore (e.g. by
 * {@link IContextStore#registerHandler(IContextHandler)}).
 * <p>
 * Then, receiving events is accomplished through the {@link #handleContextEvent(ContextEvent)} method, whereas creating new context events is achieved by
 * calling {@link IContextStore#newContextEvent(String, String, Object, java.util.Date)}.
 * <p>
 * <i>OSGi notice:</i> In OSGi environments, all IContextHandlers which are registered in the service registry will be automatically registered in the
 * ContextStore as well.
 */
public interface IContextHandler {

	/**
	 * This method is necessary for the ContextStore service tracking mechanism which tracks all registered {@link IContextHandler}s: the returned string array
	 * should contain the types of the context events which this handler should receive after (OSGi service) registering.
	 * <p>
	 * However, the types handled by a context handler may be modified during runtime, e.g. by using
	 * {@link IContextStore#registerHandlerForContextType(IContextHandler, String)}.
	 * <p>
	 * If the handler acts as a producer only, a <code>null</code>-returning dummy implementation suffices.
	 */
	public String[] getDefaultContextTypes();

	/**
	 * This method is necessary for the ContextStore service tracking mechanism which tracks all registered {@link IContextHandler}s: by passing the reference
	 * to the context store as a parameter, the handler obtains access to the context store (and should store the reference as a local member variable for
	 * further use)
	 * 
	 * @param contextStore
	 *            the context store
	 */
	public void setContextStore(IContextStore contextStore);

	/**
	 * This method processes received context events. If the handler acts as a producer only, an empty dummy implementation suffices.
	 * <p>
	 * Since version 1.1.0, the handling logic of context handlers (i.e. this method) is executed in separate context handling worker threads, thereby
	 * decoupling the handling execution time from the rest of the system. Hence, it is not necessary (and not advisable) to implement own threading in this
	 * method.
	 * 
	 * @param context
	 *            the context event to be handled
	 */
	public void handleContextEvent(ContextEvent context);
}
