/**
 * (c) Atos IT Solutions and Services GmbH, 2011.
 * Licensed under EUPL 1.1
 */
package com.sag.osami.contextstore;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.atomic.AtomicLong;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.sag.osami.api.context.ContextEvent;
import com.sag.osami.api.context.IContextHandler;
import com.sag.osami.api.context.IContextQuery;
import com.sag.osami.api.context.IContextStorage;
import com.sag.osami.contextstore.osgi.OsgiContextStoreActivator;
import com.sag.osami.contextstore.serialization.ISerializationHandler;
import com.sag.osami.contextstore.serialization.ISerializedContextHandler;
import com.sag.osami.contextstore.serialization.ISerializedContextReceiver;
import com.sag.osami.contextstore.storage.ISerializedContextStorage;
import com.sag.osami.contextstore.util.NamedThreadFactory;

/**
 * Being the actual implementation of the ContextStore pattern, this class manages {@link IContextHandler}s and dispatches and persists {@link ContextEvent}s.
 * <p>
 * It implements all generic functionality, in particular handler-specific functionality as (de)registering context handlers and passing over context events to
 * the according (interested) handlers.
 * <p>
 * It delegates all storage-related methods as loading, storing and querying {@link ContextEvent}s to an implementation of a {@link ISerializedContextStorage}.
 * <p>
 * It is noteworthy that this version of the ContextStore uses worker threads for the processing of the registered {@link IContextHandler}s. Thus the semantics
 * of context event creation changed crucially: <strong> {@link #newContextEvent(String, String, Object, Date)} etc. are no longer synchronous</strong>, i.e. it
 * cannot be assumed that the newly created event is already handled by all interested handlers after <code>newContextEvent</code> was executed.
 * <p>
 * By processing the context handlers concurrently, the overall speed is improved, especially on modern multi-core CPU architectures (in fact, the number of
 * handler worker threads corresponds to the number of CPU cores). Methods like {@link #registerHandler(IContextHandler)} etc. are thread-safe.
 * <p>
 * For creating a ContextStore instance conveniently, the builder class {@link ContextStoreBuilder} can be used.
 * <p>
 * All internally used components such as the serialization handler and the storage implementation are used strictly through well-defined interfaces and
 * therefore can be replaced with custom implementations. See {@link ContextStoreBuilder} and {@link OsgiContextStoreActivator} for an overview how to replace
 * the default components.
 * 
 * @see ISerializedContextStorage
 * @see IContextHandler
 * @see ContextStoreBuilder
 * @see OsgiContextStoreActivator
 */
public final class ContextStore implements ISerializedContextReceiver, IContextQuery {

	private final Map<IContextHandler, RegisteredHandler> registeredHandlers;
	private final Set<IContextHandler> deactivatedHandlers;
	private final List<String> registeredContextTypes;

	private final ExecutorService handlerWorkers;

	private final Logger log;
	private final ISerializedContextStorage storage;
	private final AtomicLong lastID;
	private final ISerializationHandler serializer;

	private final String storeID;

	public ContextStore(ISerializationHandler serializer, ISerializedContextStorage storage, String storeID) {
		if (storeID == null)
			throw new NullPointerException("ID of a ContextStore must not be null");
		if (storage == null)
			throw new NullPointerException("The storage handler of a ContextStore must not be null");
		if (serializer == null)
			throw new NullPointerException("The serializer handler of a ContextStore must not be null");

		log = LoggerFactory.getLogger(getClass());

		this.storeID = storeID;
		this.storage = storage;
		this.serializer = serializer;

		storage.setSerializer(serializer);

		registeredHandlers = Collections.synchronizedMap(new HashMap<IContextHandler, RegisteredHandler>());
		deactivatedHandlers = Collections.synchronizedSet(new HashSet<IContextHandler>());
		registeredContextTypes = Collections.synchronizedList(new ArrayList<String>());

		// malicious code in ContextHandlers CAN lead to thread contention when the pool is fixed, but simply
		// creating more threads isn't a solution as well - hence, the thread pool is fixed
		handlerWorkers = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors(), new NamedThreadFactory("ContextHandler-Worker-", true));

		lastID = new AtomicLong(storage.retrieveLastID());

		registeredContextTypes.add(TYPE_REGISTERED);
		registeredContextTypes.add(TYPE_DEREGISTERED);
	}

	@Override
	public void registerHandler(IContextHandler contextHandler, String[] typeList) {
		contextHandler.setContextStore(this);
		RegisteredHandler regh = null;

		if ((regh = registeredHandlers.get(contextHandler)) != null)
			regh.addTypes(typeList);
		else
			registeredHandlers.put(contextHandler, new RegisteredHandler(contextHandler, typeList));

		log.debug("registerHandler(" + contextHandler + ", " + Arrays.toString(typeList) + ")");
	}

	@Override
	public void registerHandler(IContextHandler contextHandler) {
		registerHandler(contextHandler, contextHandler.getDefaultContextTypes());
	}

	@Override
	public void registerHandlerForContextType(IContextHandler handler, String contextType) {
		registerHandler(handler, new String[] { contextType });
	}

	@Override
	public void deregisterHandler(IContextHandler contextHandler) {
		registeredHandlers.remove(contextHandler);
		log.debug("deregisterHandler(" + contextHandler + ")");
	}

	@Override
	public void newContextEvent(ContextEvent cev) {
		newContextEvent(cev, true);
	}

	@Override
	public void newContextEvent(ContextEvent cev, boolean persist) {
		// simply call the big routine.. with the components
		newContextEvent(cev.getOriginID(), cev.getType(), cev.getData(), cev.getTimestamp(), persist);
	}

	@Override
	public void newContextEvent(String originID, String type, Object data, Date date) {
		newContextEvent(originID, type, data, date, true);
	}

	@Override
	public void newContextEvent(String originID, String type, Object data, Date date, boolean persist) {
		log.debug("newContextEvent(origin={}, type={}, date={})", new Object[] { originID, type, date });

		Long id = persist ? lastID.incrementAndGet() : null;
		ContextEvent event = new ContextEvent(id, data, originID, date, type, getStoreID());

		// let the context handlers work as soon as possible!
		sendContextEventToAllRegisteredHandlers(event, false);

		if (persist) {
			String[] serializedData = serializer.serialize(data);
			ContextEvent serializedEvent = new ContextEvent(id, serializedData, originID, date, type, getStoreID());

			// again, as soon as possible ...
			sendContextEventToAllRegisteredHandlers(serializedEvent, true);

			// and now the I/O-heavy database stuff
			storage.store(false, serializedEvent);
		}
	}

	@Override
	public void newContextEvent(String originID, String type, String[] data, Date date, String contextStoreOriginID) {
		newContextEvent(originID, type, data, date, contextStoreOriginID, true);
	}

	@Override
	public void newContextEvent(String originID, String type, String[] data, Date date, String contextStoreOriginID, boolean persist) {
		// reversed order of handlers (compared to regular #newContextEvent)
		log.debug("newContextEvent(origin={}, type={}, date={}, csOrigin={}, size={})", new Object[] { originID, type, date, contextStoreOriginID,
				(data == null ? 0 : data[1].length()) });

		long id = persist ? lastID.incrementAndGet() : null;
		ContextEvent serializedEvent = new ContextEvent(id, data, originID, date, type, contextStoreOriginID);

		sendContextEventToAllRegisteredHandlers(serializedEvent, true);

		try {
			Object objData = serializer.deserialize(data);
			ContextEvent event = new ContextEvent(id, objData, originID, date, type, contextStoreOriginID);
			sendContextEventToAllRegisteredHandlers(event, false);
		} catch (Exception e) {
			log.error("Deserialization of an external ContextEvent failed!", e);
		}

		if (persist)
			storage.store(false, serializedEvent);
	}

	@Override
	public ContextEvent[] query(String query, Object... params) {
		return storage.query(query, params);
	}

	@Override
	public Iterable<ContextEvent> queryLazy(String query, Object... params) throws UnsupportedOperationException {
		return storage.queryLazy(query, params);
	}

	@Override
	public void activateEventsForHandler(IContextHandler myHandler) {
		deactivatedHandlers.remove(myHandler);
		log.debug("activateEventsForHandler(" + myHandler + ")");
	}

	@Override
	public void deactivateEventsForHandler(IContextHandler myHandler) {
		deactivatedHandlers.add(myHandler);
		log.debug("deactivateEventsForHandler(" + myHandler + ")");
	}

	@Override
	public void deregisterContextType(String myType) {
		if (registeredContextTypes.contains(myType)) {
			log.debug("deregisterContextType(" + myType + ")");
			registeredContextTypes.remove(myType);
			newContextEvent(CONTEXT_STORE_ID, TYPE_DEREGISTERED, myType, new Date(), false);
		}
	}

	@Override
	public boolean isRegisteredContextType(String myType) {
		return registeredContextTypes.contains(myType);
	}

	@Override
	public String[] getRegisteredContextTypes() {
		return registeredContextTypes.toArray(new String[0]);
	}

	@Override
	public void registerContextType(String myType) {
		if (!registeredContextTypes.contains(myType)) {
			log.debug("registerContextType(" + myType + ")");
			registeredContextTypes.add(myType);
			newContextEvent(CONTEXT_STORE_ID, TYPE_REGISTERED, myType, new Date(), false);
		}
	}

	@Override
	public void deregisterHandlerForContextType(IContextHandler handler, String contextType) {
		registeredHandlers.get(handler).removeType(contextType);
		log.debug(String.format("deregisterHandlerForContextType(%s, %s)", handler, contextType));
	}

	private void sendContextEventToAllRegisteredHandlers(final ContextEvent contextEvent, boolean isSerializedEvent) {
		List<IContextHandler> concernedHandlers = new LinkedList<IContextHandler>();
		String contextType = contextEvent.getType();

		synchronized (registeredHandlers) {
			for (RegisteredHandler handler : registeredHandlers.values()) {
				boolean isSerializedHandler = handler.getContextHandler() instanceof ISerializedContextHandler;
				if (isSerializedEvent == isSerializedHandler)
					if (handler.listensOnType(contextType))
						if (!deactivatedHandlers.contains(handler))
							concernedHandlers.add(handler.getContextHandler());
			}
		}

		if (!registeredContextTypes.contains(contextType))
			registerContextType(contextType);

		for (final IContextHandler handler : concernedHandlers) {
			handlerWorkers.execute(new Runnable() {
				public void run() {
					handler.handleContextEvent(contextEvent);
				}
			});
		}
	}

	@Override
	public String[] getHandlersRegisteredContextTypes(IContextHandler handler) {
		return registeredHandlers.get(handler).getTypes();
	}

	@Override
	public String getStoreID() {
		return storeID;
	}

	public ISerializationHandler getSerializationHandler() {
		return serializer;
	}

	@Override
	public IContextStorage getStorage() {
		return storage;
	}

	/** Wrapper class for managing the context types for which a handler is currently registered */
	private static class RegisteredHandler {

		IContextHandler contextHandler = null;
		Set<String> typeList;

		RegisteredHandler(IContextHandler contextHandler, String[] typeList) {
			this.contextHandler = contextHandler;
			this.typeList = Collections.synchronizedSet(new LinkedHashSet<String>((typeList != null) ? typeList.length : 16));
			addTypes(typeList);
		}

		IContextHandler getContextHandler() {
			return contextHandler;
		}

		boolean listensOnType(String type) {
			return typeList.contains(type) || typeList.contains(TYPE_ALL);
		}

		String[] getTypes() {
			return typeList.toArray(new String[typeList.size()]);
		}

		void addTypes(String[] typeList) {
			if (typeList != null)
				for (String type : typeList)
					this.typeList.add(type);
		}

		void removeType(String type) {
			typeList.remove(type);
		}
	}
}