/**
 * (c) Atos IT Solutions and Services GmbH, 2011.
 * Licensed under EUPL 1.1
 */
package com.sag.osami.contextstore;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.InetAddress;
import java.util.Properties;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.sag.osami.api.context.ContextEvent;
import com.sag.osami.contextstore.osgi.OsgiContextStoreActivator;
import com.sag.osami.contextstore.serialization.ISerializationHandler;
import com.sag.osami.contextstore.serialization.json.JSONSerializer;
import com.sag.osami.contextstore.storage.ISerializedContextStorage;
import com.sag.osami.contextstore.storage.impl.JDBCStorage;
import com.sag.osami.contextstore.storage.jdbc.DerbyDialect;
import com.sag.osami.contextstore.storage.jdbc.IJDBCDialect;

/**
 * This builder class can be used in non-OSGi-evnironments to configure and create a ContextStore instance conveniently.
 * <p>
 * It is also used internally in the {@link OsgiContextStoreActivator} (using registered OSGi services as the respective implementations).
 * <p>
 * 
 * It provides methods (all of which return the builder itself, allowing to chain the configuration in one line) for configuring:
 * 
 * <p>
 * <ul>
 * <li><code>{@link #storeID(String)}</code>
 * <li><code>{@link #serializer(ISerializationHandler)}</code>
 * <li><code>{@link #storage(ISerializedContextStorage)}</code>
 * <li><code>{@link #jdbcDialect(IJDBCDialect)}</code>
 * </ul>
 * 
 * Additionally, it can be configured via a .properties configuration file {@value #PROPERTIES_FILENAME} in '/' on the CLASSPATH or any file defined via the
 * builder constructor. Also, it can be configured via system properties (using the same configuration keys as in the .properties file, but with the prefix
 * {@value #SYSTEM_PROPERTY_PREFIX}).
 * 
 * <p>
 * Following properties can be configured:
 * 
 * <ul>
 * <li><code>{@value #CONFIG_STORE_ID}</code>: the context store ID
 * <li>see {@link JDBCStorage} for JDBC-related properties
 * </ul>
 * 
 * The most simple, but working usage example for this class looks like this:
 * 
 * <pre>
 * <code>ContextStore store = new ContextStoreBuilder().build();</code>
 * </pre>
 */
public class ContextStoreBuilder {

	private static final String PROPERTIES_FILENAME = "contextstore.properties";
	private static final String SYSTEM_PROPERTY_PREFIX = "com.sag.osami.contextstore.";

	public static final String CONFIG_STORE_ID = "contextstore_id";

	private Properties properties;
	private ISerializationHandler serializer;
	private ISerializedContextStorage storage;
	private IJDBCDialect jdbcDialect;

	private static Logger log = LoggerFactory.getLogger(ContextStoreBuilder.class);

	private static Properties defaultsProperties() {
		Properties defaults = new Properties();
		try {
			// DEFAULT STORE ID: hostname
			defaults.setProperty(CONFIG_STORE_ID, InetAddress.getLocalHost().getHostName());
			// DEFAULT JDBC URL: Derby
			defaults.setProperty(JDBCStorage.CONFIG_URL, DerbyDialect.DEFAULT_URL);
		} catch (Exception e) {
			// does not happen
			e.printStackTrace();
		}
		return defaults;
	}

	private void setDefaultComponents() {
		// DEFAULT SERIALIZER: JSON
		if (serializer == null) {
			log.info("Using JSONSerializer as ContextStore serialization handler by default");
			serializer = new JSONSerializer();
		}

		// DEFAULT STORAGE: JDBC
		if (storage == null) {
			log.info("Using JDBCStorage as ContextStore storage engine by default");
			if (jdbcDialect != null) {
				storage = new JDBCStorage(jdbcDialect, properties);
			} else {
				// figure out dialect automatically (using JDBC URL)
				storage = new JDBCStorage(properties);
			}
		}
	}

	private boolean loadProperties(InputStream in) {
		try {
			properties.load(in);
			return true;
		} catch (Exception e) {
			return false;
			// never mind
		} finally {
			try {
				if (in != null)
					in.close();
			} catch (IOException e) {
				// shouldn't happen
			}
		}
	}

	private void loadSystemProperties() {
		for (Object _sysprop : System.getProperties().keySet()) {
			String sysprop = (String) _sysprop;
			if (sysprop.startsWith(SYSTEM_PROPERTY_PREFIX)) {
				String shortProp = sysprop.substring(SYSTEM_PROPERTY_PREFIX.length());
				properties.setProperty(shortProp, System.getProperty(sysprop));
				log.info("ContextStore property loaded from system properties: {}", sysprop);
			}
		}
	}

	/**
	 * Creates a new ContextStore builder, overriding defaults with the given properties.
	 * 
	 * @param props
	 *            configuration properties
	 */
	public ContextStoreBuilder(Properties props) {
		initialize();

		// properties from constructor
		if (props != null)
			for (Object pk : props.keySet())
				properties.setProperty((String) pk, props.getProperty((String) pk));
	}

	/**
	 * Creates a new ContextStore builder, overriding defaults with the properties from the given file.
	 * 
	 * @param propsFile
	 *            '.properties' file
	 */
	public ContextStoreBuilder(File propsFile) {
		initialize();

		// properties file from constructor
		try {
			loadProperties(new FileInputStream(propsFile));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Creates a new ContextStore builder with defaults.
	 */
	public ContextStoreBuilder() {
		initialize();
	}

	private void initialize() {
		properties = new Properties(defaultsProperties());

		// .properties from current working dir
		try {
			File propsFile = new File(PROPERTIES_FILENAME);
			if (loadProperties(new FileInputStream(propsFile)))
				log.info("ContextStore properties loaded: {}", propsFile.getAbsolutePath());
		} catch (FileNotFoundException e) {
			// never mind
		}

		// .properties from CLASSPATH
		if (loadProperties(getClass().getResourceAsStream("/" + PROPERTIES_FILENAME)))
			log.info("ContextStore propreties loaded from class path");

		// system properties
		loadSystemProperties();
	}

	/**
	 * Sets the serialization handler implementation, which is used for serializing and deserializing context events (e.g. when persisting).
	 * <p>
	 * Default: {@link JSONSerializer}
	 * 
	 * @param serializer
	 *            the implementation of the serialization handler
	 * @return the builder itself
	 */
	public ContextStoreBuilder serializer(ISerializationHandler serializer) {
		if (serializer == null)
			throw new NullPointerException();

		this.serializer = serializer;
		log.info("Using {} as ContextStore serialization handler", serializer.getClass().getName());
		return this;
	}

	/**
	 * Sets the storage implementation, which is responsible for persisting, loading and querying context events.
	 * <p>
	 * Default: {@link JDBCStorage} with {@link DerbyDialect} and {@link DerbyDialect#DEFAULT_URL}.
	 * 
	 * @param storage
	 *            the implementation of the storage
	 * @return the builder itself
	 * @see #jdbcDialect(IJDBCDialect)
	 */
	public ContextStoreBuilder storage(ISerializedContextStorage storage) {
		if (storage == null)
			throw new NullPointerException();

		this.storage = storage;
		log.info("Using {} as ContextStore storage engine", storage.getClass().getName());
		return this;
	}

	/**
	 * Sets the dialect to be used by the {@link JDBCStorage}.
	 * <p>
	 * Default: {@link DerbyDialect}
	 * 
	 * @param jdbcDialect
	 *            the implementation of the JDBC dialect
	 * @return the builder itself
	 */
	public ContextStoreBuilder jdbcDialect(IJDBCDialect jdbcDialect) {
		if (jdbcDialect == null)
			throw new NullPointerException();

		this.jdbcDialect = jdbcDialect;
		log.info("Using {} as JDBCStorage dialect", jdbcDialect.getClass().getName());
		return this;
	}

	/**
	 * Sets the context store ID as used in {@link ContextEvent#getContextStoreOriginID()}.
	 * <p>
	 * Default: hostname of the executing machine
	 * 
	 * @param storeID
	 *            the context store ID
	 * @return the builder itself
	 */
	public ContextStoreBuilder storeID(String storeID) {
		if (storeID == null)
			throw new NullPointerException();

		properties.setProperty(CONFIG_STORE_ID, storeID);
		log.info("Using '{}' as ContextStore ID", storeID);
		return this;
	}

	/**
	 * Builds the ContextStore according to the configuration.
	 * 
	 * @return the context store instance
	 */
	public ContextStore build() {
		setDefaultComponents();
		return new ContextStore(serializer, storage, properties.getProperty(CONFIG_STORE_ID));
	}
}