/**
 * (c) Atos IT Solutions and Services GmbH, 2011.
 * Licensed under EUPL 1.1
 */
package com.sag.osami.contextstore.storage;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;

/**
 * This interface defines table name and column name constants for a database-backed ContextStore storage implementation. Storage implementation classes should
 * use this constants so that e.g. query writers know how to refer to columns etc.
 * <p>
 * A context store table based on the constants defined in this interface would look like this:
 * <p>
 * <table border=1>
 * <tr>
 * <th colspan=2>{@value #TABLE_NAME}</th>
 * </tr>
 * <tr>
 * <th>Column name</th>
 * <th>Example value</th>
 * </tr>
 * <tr>
 * <td>{@value #COL_ID}</td>
 * <td>auto incremented</td>
 * </tr>
 * <tr>
 * <td>{@value #COL_ORIGINID}</td>
 * <td>COM/SAG/OSAMI/temp1</td>
 * </tr>
 * <tr>
 * <td>{@value #COL_TYPE}</td>
 * <td>TEMP_C</td>
 * </tr>
 * <tr>
 * <td>{@value #COL_DATA_CLASS}</td>
 * <td>java.lang.Long</td>
 * </tr>
 * <tr>
 * <td>{@value #COL_DATA}</td>
 * <td>(serialized payload data, e.g. JSON-encoded)</td>
 * </tr>
 * <tr>
 * <td>{@value #COL_CTXSTORE_ID}</td>
 * <td>the-hostname</td>
 * </tr>
 * <tr>
 * <td>{@value #COL_TIMESTAMP}</td>
 * <td>2011-12-24 00:00:01.984</td>
 * </tr>
 * </table>
 */
public interface DBConstants {

	static final String TABLE_NAME = "contextstore";

	// columns
	static final String COL_ID = "id";
	static final String COL_ORIGINID = "originID";
	static final String COL_TYPE = "type";
	static final String COL_DATA = "data";
	static final String COL_DATA_CLASS = "dataClass";
	static final String COL_CTXSTORE_ID = "contextStoreOriginID";
	static final String COL_TIMESTAMP = "timestamp";

	/**
	 * An unmodifiable list containing the column names of the ContextStore table. This is a convenience constant. The order of the columns is:
	 * 
	 * <ol>
	 * <li>id
	 * <li>originID
	 * <li>type
	 * <li>dataClass
	 * <li>data
	 * <li>contextStoreOriginID
	 * <li>timestamp
	 * </ol>
	 */
	static final List<String> COLS = Collections.unmodifiableList(Arrays.asList(COL_ID, COL_ORIGINID, COL_TYPE, COL_DATA_CLASS, COL_DATA, COL_CTXSTORE_ID,
			COL_TIMESTAMP));

	/**
	 * An unmodifiable list containing the table name and colum names of the ContextStore table. This is a convenience constant. The order is:
	 * 
	 * <ol>
	 * <li>contextstore [table name]
	 * <li>id
	 * <li>originID
	 * <li>type
	 * <li>dataClass
	 * <li>data
	 * <li>contextStoreOriginID
	 * <li>timestamp
	 * </ol>
	 * 
	 */
	static final List<String> TABLE_AND_COLS = Collections.unmodifiableList(Arrays.asList(TABLE_NAME, COL_ID, COL_ORIGINID, COL_TYPE, COL_DATA_CLASS, COL_DATA,
			COL_CTXSTORE_ID, COL_TIMESTAMP));
}