/**
 * (c) Atos IT Solutions and Services GmbH, 2011.
 * Licensed under EUPL 1.1
 */
package com.sag.osami.contextstore.storage.impl;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.RandomAccess;

import org.slf4j.Logger;

import com.sag.osami.api.context.ContextEvent;

/**
 * Convenience methods and classes for handling iterators
 */
final class IteratorUtil {

	private IteratorUtil() {
	}

	/**
	 * Stores the values returned by an iterator in a list
	 * 
	 * @param iter
	 *            the iterator
	 * @return list with the values returned by the iterator
	 */
	static List<ContextEvent> toList(Iterator<ContextEvent> iter) {
		List<ContextEvent> list = new ArrayList<ContextEvent>();
		while (iter.hasNext())
			list.add(iter.next());
		return list;
	}

	/**
	 * Stores the values returned by an iterator in an array
	 * 
	 * @param iter
	 *            the iterator
	 * @return array with the values returned by the iterator
	 */
	static ContextEvent[] toArray(Iterator<ContextEvent> iter) {
		List<ContextEvent> list = toList(iter);
		return (ContextEvent[]) list.toArray(new ContextEvent[list.size()]);
	}

	/**
	 * Abstract base class for ContextEvent iterators which receive their elements from a cursor-based database library, e.g. JDBC or the Android SQLite API.
	 */
	static abstract class RowIterator implements Iterator<ContextEvent> {

		protected static class IteratorException extends Exception {
			private static final long serialVersionUID = 1L;

			public IteratorException(Exception cause) {
				super(cause);
			}
		}

		protected final boolean deserialize;
		protected final String query;
		protected final Object[] params;
		protected final Logger log;

		private boolean initialized, closed;

		private ContextEvent curEvent, nextEvent;

		protected abstract void initialize() throws IteratorException;

		protected abstract boolean moveNext() throws IteratorException;

		protected abstract ContextEvent loadContextEvent() throws IteratorException;

		protected abstract void close();

		protected RowIterator(Logger log, boolean deserialize, String query, Object... params) {
			this.log = log;
			this.deserialize = deserialize;
			this.query = query;
			this.params = params;
		}

		protected final void logErrorAndClose(IteratorException e) {
			log.error("ContextStore query failed", e);
			wrappedClose();
		}

		private void wrappedClose() {
			if (!closed) {
				closed = true;
				close();
			}
		}

		private void wrappedInitialize() {
			if (!initialized)
				try {
					initialize();
					initialized = true;
				} catch (IteratorException e) {
					logErrorAndClose(e);
				}
		}

		@Override
		public final boolean hasNext() {
			if (curEvent != null)
				return true;

			if (curEvent == null && nextEvent != null) {
				curEvent = nextEvent;
				nextEvent = null;
				return true;
			}

			if (!closed && curEvent == null && nextEvent == null) {
				wrappedInitialize();

				// move to next (& close if we are past the last row)
				try {
					if (!moveNext()) {
						wrappedClose();
						return false;
					}

					// cache the current ContextEvent
					curEvent = loadContextEvent();
				} catch (IteratorException e) {
					logErrorAndClose(e);
					return false;
				}

				// check if we still have a next row (and cache the next ContextEvent as well)
				try {
					if (moveNext()) {
						nextEvent = loadContextEvent();
					} else {
						wrappedClose();
					}
				} catch (IteratorException e) {
					logErrorAndClose(e);
				}

				// but we still have a current event which was not returned via next() yet
				return true;
			}

			return false;
		}

		@Override
		public final ContextEvent next() {
			hasNext();
			if (curEvent != null) {
				ContextEvent cur = curEvent;
				curEvent = null;
				return cur;
			}
			throw new NoSuchElementException();
		}

		@Override
		public final void remove() {
			throw new UnsupportedOperationException();
		}
	}

	/**
	 * Transparently chains several iterators into one iterator.
	 * 
	 * @param <T>
	 *            type of the iterator
	 */
	static class ChainedIterator<T> implements Iterator<T> {

		private final List<Iterator<T>> chain;
		private int cur;

		public ChainedIterator(List<Iterator<T>> chain) {
			this.cur = 0;

			if (chain instanceof RandomAccess)
				this.chain = chain;
			else
				this.chain = new ArrayList<Iterator<T>>(chain);
		}

		@Override
		public boolean hasNext() {
			while (cur < chain.size()) {
				if (chain.get(cur).hasNext())
					return true;
				else
					cur++;
			}
			return false;
		}

		@Override
		public T next() {
			while (cur < chain.size()) {
				try {
					return chain.get(cur).next();
				} catch (NoSuchElementException e) {
					cur++;
				}
			}
			throw new NoSuchElementException();
		}

		@Override
		public void remove() {
			chain.get(cur).remove();
		}
	}
}