/**
 * (c) Atos IT Solutions and Services GmbH, 2011.
 * Licensed under EUPL 1.1
 */
package com.sag.osami.contextstore.storage.jdbc;

import com.sag.osami.contextstore.storage.DBConstants;
import com.sag.osami.contextstore.storage.impl.JDBCStorage;

/**
 * A basic implementation of a JDBC dialect, using only standard SQL.
 * <p>
 * However, as column type naming and behavior is often very database-specific, this aspect is left to actual dialect implementations. In particular,
 * {@link #createTable()} remains unimplemented, but the convenience method {@link #createTable(String[])} is provided.
 * <p>
 * Naturally, {@link #driver()} also remains unimplemented.
 * 
 * @see JDBCStorage
 * @see IJDBCDialect
 */
public abstract class GenericJDBCDialect implements IJDBCDialect {

	private static final String TABLE_EXISTS = "SELECT * FROM " + TABLE_NAME + " WHERE 1 = 0";
	private static final String INSERT_CONTEXT_EVENT = String.format("INSERT INTO %s (%s,%s,%s,%s,%s,%s,%s) VALUES(?,?,?,?,?,?,?)", TABLE_AND_COLS.toArray());
	private static final String LOAD_CONTEXT_EVENT = "SELECT * FROM " + TABLE_NAME + " WHERE " + COL_ID + " = ?";
	private static final String DELETE_CONTEXT_EVENT = "DELETE FROM " + TABLE_NAME + " WHERE " + COL_ID + " = ?";
	private static final String QUERY_LAST_ID = "SELECT MAX(" + COL_ID + ") FROM " + TABLE_NAME;

	public String tableExists() {
		return TABLE_EXISTS;
	}

	/**
	 * Returns a basic CREATE TABLE statement, using the {@link DBConstants#TABLE_AND_COLS} list for the table and column names and the colTypes parameter for
	 * the column types.
	 * 
	 * @param colTypes
	 *            the data types of the columns. The order within the array should match the one in {@link DBConstants#COLS}
	 * @return the CREATE TABLE SQL statement
	 */
	protected String createTable(String[] colTypes) {
		String[] cols = COLS.toArray(new String[COLS.size()]);
		StringBuilder createTable = new StringBuilder("CREATE TABLE " + TABLE_NAME + "(");

		for (int i = 0; i < cols.length; i++)
			createTable.append(String.format("%s %s %s", (i == 0 ? "" : ","), cols[i], colTypes[i]));

		createTable.append(")");
		return createTable.toString();
	}

	@Override
	public String dropTable() {
		return "DROP TABLE " + TABLE_NAME;
	}

	@Override
	public String insertContextEvent() {
		return INSERT_CONTEXT_EVENT;
	}

	@Override
	public String loadContextEvent() {
		return LOAD_CONTEXT_EVENT;
	}

	@Override
	public String deleteContextEvent() {
		return DELETE_CONTEXT_EVENT;
	}

	@Override
	public String loadContextEvents(int idsCount) throws UnsupportedOperationException {
		return "SELECT * FROM " + TABLE_NAME + " WHERE " + COL_ID + " IN (" + idParams(idsCount) + ")";
	}

	@Override
	public String queryLastID() {
		return QUERY_LAST_ID;
	}

	/**
	 * Convenience method for generating a (?,?,?,...,?) string to be used in an SQL query, especially in {@link #loadContextEvents(int)}
	 * 
	 * @param idsCount
	 *            count of the parameters
	 * @return (?,?,?,...,?)
	 */
	protected static String idParams(int idsCount) {
		StringBuilder idParams = new StringBuilder(idsCount * 2 - 1);
		for (int i = 0; i < idsCount; i++)
			idParams.append(i == 0 ? "?" : ",?");
		return idParams.toString();
	}
}